@@ -41,9 +41,11 @@ def write_commit_patch(f, commit, contents, progress, version=None,
                        encoding=None):
     """Write a individual file patch.
 
-    :param commit: Commit object
-    :param progress: Tuple with current patch number and total.
-    :return: tuple with filename and contents
+    Args:
+      commit: Commit object
+      progress: Tuple with current patch number and total.
+    Returns:
+      tuple with filename and contents
     """
     encoding = encoding or getattr(f, "encoding", "ascii")
     if isinstance(contents, str):
@@ -80,10 +82,12 @@ def write_commit_patch(f, commit, contents, progress, version=None,
 def get_summary(commit):
     """Determine the summary line for use in a filename.
 
-    :param commit: Commit
-    :return: Summary string
+    Args:
+      commit: Commit
+    Returns: Summary string
     """
-    return commit.message.splitlines()[0].replace(" ", "-")
+    decoded = commit.message.decode(errors='replace')
+    return decoded.splitlines()[0].replace(" ", "-")
 
 
 #  Unified Diff
@@ -100,7 +104,8 @@ def _format_range_unified(start, stop):
 
 
 def unified_diff(a, b, fromfile='', tofile='', fromfiledate='',
-                 tofiledate='', n=3, lineterm='\n'):
+                 tofiledate='', n=3, lineterm='\n', tree_encoding='utf-8',
+                 output_encoding='utf-8'):
     """difflib.unified_diff that can detect "No newline at end of file" as
     original "git diff" does.
 
@@ -113,15 +118,15 @@ def unified_diff(a, b, fromfile='', tofile='', fromfiledate='',
             fromdate = '\t{}'.format(fromfiledate) if fromfiledate else ''
             todate = '\t{}'.format(tofiledate) if tofiledate else ''
             yield '--- {}{}{}'.format(
-                fromfile.decode("ascii"),
+                fromfile.decode(tree_encoding),
                 fromdate,
                 lineterm
-                ).encode('ascii')
+                ).encode(output_encoding)
             yield '+++ {}{}{}'.format(
-                tofile.decode("ascii"),
+                tofile.decode(tree_encoding),
                 todate,
                 lineterm
-                ).encode('ascii')
+                ).encode(output_encoding)
 
         first, last = group[0], group[-1]
         file1_range = _format_range_unified(first[1], last[2])
@@ -130,7 +135,7 @@ def unified_diff(a, b, fromfile='', tofile='', fromfiledate='',
             file1_range,
             file2_range,
             lineterm
-             ).encode('ascii')
+             ).encode(output_encoding)
 
         for tag, i1, i2, j1, j2 in group:
             if tag == 'equal':
@@ -152,7 +157,8 @@ def unified_diff(a, b, fromfile='', tofile='', fromfiledate='',
 def is_binary(content):
     """See if the first few bytes contain any null characters.
 
-    :param content: Bytestring to check for binary content
+    Args:
+      content: Bytestring to check for binary content
     """
     return b'\0' in content[:FIRST_FEW_BYTES]
 
@@ -174,25 +180,26 @@ def patch_filename(p, root):
 def write_object_diff(f, store, old_file, new_file, diff_binary=False):
     """Write the diff for an object.
 
-    :param f: File-like object to write to
-    :param store: Store to retrieve objects from, if necessary
-    :param old_file: (path, mode, hexsha) tuple
-    :param new_file: (path, mode, hexsha) tuple
-    :param diff_binary: Whether to diff files even if they
+    Args:
+      f: File-like object to write to
+      store: Store to retrieve objects from, if necessary
+      old_file: (path, mode, hexsha) tuple
+      new_file: (path, mode, hexsha) tuple
+      diff_binary: Whether to diff files even if they
         are considered binary files by is_binary().
 
-    :note: the tuple elements should be None for nonexistant files
+    Note: the tuple elements should be None for nonexistant files
     """
     (old_path, old_mode, old_id) = old_file
     (new_path, new_mode, new_id) = new_file
-    old_path = patch_filename(old_path, b"a")
-    new_path = patch_filename(new_path, b"b")
+    patched_old_path = patch_filename(old_path, b"a")
+    patched_new_path = patch_filename(new_path, b"b")
 
     def content(mode, hexsha):
         if hexsha is None:
             return Blob.from_string(b'')
         elif S_ISGITLINK(mode):
-            return Blob.from_string(b"Submodule commit " + hexsha + b"\n")
+            return Blob.from_string(b"Subproject commit " + hexsha + b"\n")
         else:
             return store[hexsha]
 
@@ -207,34 +214,48 @@ def write_object_diff(f, store, old_file, new_file, diff_binary=False):
     new_content = content(new_mode, new_id)
     if not diff_binary and (
             is_binary(old_content.data) or is_binary(new_content.data)):
-        f.write(b"Binary files " + old_path + b" and " + new_path +
-                b" differ\n")
+        binary_diff = (
+            b"Binary files "
+            + patched_old_path
+            + b" and "
+            + patched_new_path
+            + b" differ\n"
+        )
+        f.write(binary_diff)
     else:
         f.writelines(unified_diff(lines(old_content), lines(new_content),
-                     old_path, new_path))
+                     patched_old_path, patched_new_path))
 
 
 # TODO(jelmer): Support writing unicode, rather than bytes.
 def gen_diff_header(paths, modes, shas):
     """Write a blob diff header.
 
-    :param paths: Tuple with old and new path
-    :param modes: Tuple with old and new modes
-    :param shas: Tuple with old and new shas
+    Args:
+      paths: Tuple with old and new path
+      modes: Tuple with old and new modes
+      shas: Tuple with old and new shas
     """
     (old_path, new_path) = paths
     (old_mode, new_mode) = modes
     (old_sha, new_sha) = shas
+    if old_path is None and new_path is not None:
+        old_path = new_path
+    if new_path is None and old_path is not None:
+        new_path = old_path
+    old_path = patch_filename(old_path, b"a")
+    new_path = patch_filename(new_path, b"b")
     yield b"diff --git " + old_path + b" " + new_path + b"\n"
+
     if old_mode != new_mode:
         if new_mode is not None:
             if old_mode is not None:
-                yield ("old mode %o\n" % old_mode).encode('ascii')
-            yield ("new mode %o\n" % new_mode).encode('ascii')
+                yield ("old file mode %o\n" % old_mode).encode('ascii')
+            yield ("new file mode %o\n" % new_mode).encode('ascii')
         else:
-            yield ("deleted mode %o\n" % old_mode).encode('ascii')
+            yield ("deleted file mode %o\n" % old_mode).encode('ascii')
     yield b"index " + shortid(old_sha) + b".." + shortid(new_sha)
-    if new_mode is not None:
+    if new_mode is not None and old_mode is not None:
         yield (" %o" % new_mode).encode('ascii')
     yield b"\n"
 
@@ -243,16 +264,17 @@ def gen_diff_header(paths, modes, shas):
 def write_blob_diff(f, old_file, new_file):
     """Write blob diff.
 
-    :param f: File-like object to write to
-    :param old_file: (path, mode, hexsha) tuple (None if nonexisting)
-    :param new_file: (path, mode, hexsha) tuple (None if nonexisting)
+    Args:
+      f: File-like object to write to
+      old_file: (path, mode, hexsha) tuple (None if nonexisting)
+      new_file: (path, mode, hexsha) tuple (None if nonexisting)
 
-    :note: The use of write_object_diff is recommended over this function.
+    Note: The use of write_object_diff is recommended over this function.
     """
     (old_path, old_mode, old_blob) = old_file
     (new_path, new_mode, new_blob) = new_file
-    old_path = patch_filename(old_path, b"a")
-    new_path = patch_filename(new_path, b"b")
+    patched_old_path = patch_filename(old_path, b"a")
+    patched_new_path = patch_filename(new_path, b"b")
 
     def lines(blob):
         if blob is not None:
@@ -265,16 +287,17 @@ def write_blob_diff(f, old_file, new_file):
     old_contents = lines(old_blob)
     new_contents = lines(new_blob)
     f.writelines(unified_diff(old_contents, new_contents,
-                 old_path, new_path))
+                 patched_old_path, patched_new_path))
 
 
 def write_tree_diff(f, store, old_tree, new_tree, diff_binary=False):
     """Write tree diff.
 
-    :param f: File-like object to write to.
-    :param old_tree: Old tree id
-    :param new_tree: New tree id
-    :param diff_binary: Whether to diff files even if they
+    Args:
+      f: File-like object to write to.
+      old_tree: Old tree id
+      new_tree: New tree id
+      diff_binary: Whether to diff files even if they
         are considered binary files by is_binary().
     """
     changes = store.tree_changes(old_tree, new_tree)
@@ -286,9 +309,10 @@ def write_tree_diff(f, store, old_tree, new_tree, diff_binary=False):
 def git_am_patch_split(f, encoding=None):
     """Parse a git-am-style patch and split it up into bits.
 
-    :param f: File-like object to parse
-    :param encoding: Encoding to use when creating Git objects
-    :return: Tuple with commit object, diff contents and git version
+    Args:
+      f: File-like object to parse
+      encoding: Encoding to use when creating Git objects
+    Returns: Tuple with commit object, diff contents and git version
     """
     encoding = encoding or getattr(f, "encoding", "ascii")
     encoding = encoding or "ascii"
@@ -306,9 +330,10 @@ def git_am_patch_split(f, encoding=None):
 def parse_patch_message(msg, encoding=None):
     """Extract a Commit object and patch from an e-mail message.
 
-    :param msg: An email message (email.message.Message)
-    :param encoding: Encoding to use to encode Git commits
-    :return: Tuple with commit object, diff contents and git version
+    Args:
+      msg: An email message (email.message.Message)
+      encoding: Encoding to use to encode Git commits
+    Returns: Tuple with commit object, diff contents and git version
     """
     c = Commit()
     c.author = msg["from"].encode(encoding)
