@@ -48,6 +48,7 @@ SIDE_BAND_CHANNEL_PROGRESS = 2
 # fatal error message just before stream aborts
 SIDE_BAND_CHANNEL_FATAL = 3
 
+CAPABILITY_ATOMIC = b'atomic'
 CAPABILITY_DEEPEN_SINCE = b'deepen-since'
 CAPABILITY_DEEPEN_NOT = b'deepen-not'
 CAPABILITY_DEEPEN_RELATIVE = b'deepen-relative'
@@ -66,6 +67,8 @@ CAPABILITY_SIDE_BAND_64K = b'side-band-64k'
 CAPABILITY_THIN_PACK = b'thin-pack'
 CAPABILITY_AGENT = b'agent'
 CAPABILITY_SYMREF = b'symref'
+CAPABILITY_ALLOW_TIP_SHA1_IN_WANT = b'allow-tip-sha1-in-want'
+CAPABILITY_ALLOW_REACHABLE_SHA1_IN_WANT = b'allow-reachable-sha1-in-want'
 
 # Magic ref that is used to attach capabilities to when
 # there are no refs. Should always be ste to ZERO_SHA.
@@ -87,9 +90,15 @@ KNOWN_UPLOAD_CAPABILITIES = set(COMMON_CAPABILITIES + [
     CAPABILITY_SHALLOW,
     CAPABILITY_DEEPEN_NOT,
     CAPABILITY_DEEPEN_RELATIVE,
+    CAPABILITY_ALLOW_TIP_SHA1_IN_WANT,
+    CAPABILITY_ALLOW_REACHABLE_SHA1_IN_WANT,
     ])
 KNOWN_RECEIVE_CAPABILITIES = set(COMMON_CAPABILITIES + [
-    CAPABILITY_REPORT_STATUS])
+    CAPABILITY_REPORT_STATUS,
+    CAPABILITY_DELETE_REFS,
+    CAPABILITY_QUIET,
+    CAPABILITY_ATOMIC,
+    ])
 
 
 def agent_string():
@@ -141,6 +150,17 @@ class ProtocolFile(object):
         pass
 
 
+def format_cmd_pkt(cmd, *args):
+    return cmd + b" " + b"".join([(a + b"\0") for a in args])
+
+
+def parse_cmd_pkt(line):
+    splice_at = line.find(b" ")
+    cmd, args = line[:splice_at], line[splice_at+1:]
+    assert args[-1:] == b"\x00"
+    return cmd, args[:-1].split(b"\0")
+
+
 def pkt_line(data):
     """Wrap data in a pkt-line.
 
@@ -318,7 +338,7 @@ class Protocol(object):
           cmd: The remote service to access.
           args: List of arguments to send to remove service.
         """
-        self.write_pkt_line(cmd + b" " + b"".join([(a + b"\0") for a in args]))
+        self.write_pkt_line(format_cmd_pkt(cmd, *args))
 
     def read_cmd(self):
         """Read a command and some arguments from the git client
@@ -328,10 +348,7 @@ class Protocol(object):
         Returns: A tuple of (command, [list of arguments]).
         """
         line = self.read_pkt_line()
-        splice_at = line.find(b" ")
-        cmd, args = line[:splice_at], line[splice_at+1:]
-        assert args[-1:] == b"\x00"
-        return cmd, args[:-1].split(b"\0")
+        return parse_cmd_pkt(line)
 
 
 _RBUFSIZE = 8192  # Default read buffer size.
